/*
	action: data structure for communicating orders and their execution

	part of DerMixD
	(c)2004-2008 Thomas Orgis, licensed under GPLv2
*/
#include "common.hxx"
#include "action.hxx"
#include "comparse.hxx"
#include "tstring.hxx"

const action_def* find_action_def(string &name, const action_def* world)
{
	const action_def* def = NULL;
	//find definition, slow, slow search
	int i = -1;
	if(world == NULL) return def;

	while((def == NULL) && (world[++i].id != 0))
	{
		if(name == world[i].name) def = &world[i];
	}
	#ifdef DEBUG
		if(def != NULL)
		{
			cerr << "definition for " << def->name;
			for(size_t i = 0; i < def->argc; ++i) cerr << " " << def->argt[i];
			cerr << endl;
		}
	#endif
	return def;
}

void list_actions(const action_def* world, string &app)
{
	int i = -1;
	while(world[++i].id != 0)
	{
		if(i != 0) app += " ";
		app += world[i].name;
	}
}

void action_allhelp(const action_def* world, vector<string*> &app)
{
	int i = -1;
	if(world == NULL)
	{
		app.push_back(new string("Error: No world to get action in!"));
		return;
	}
	app.push_back(new string("<name>:\t<description> <flags if any>[; parameters: parm1(type) parm2(type) ...]"));
	app.push_back(new string("meaning of flags:"));
	app.push_back(new string("\tcontainer -- this action can contain a subaction (given as parameter); read: this is the script command"));
	app.push_back(new string("\tnosub -- this action cannot be a subaction in a container; read: not allowed in script"));
	app.push_back(new string("\toptarg -- parameters are optional: either specify all or none"));
	app.push_back(new string(""));
	while(world[++i].id != 0)
	{
		app.push_back(new string);
		action_help(&world[i], *app.back());
	}
}

// append help 
void action_help(const action_def* act, string &app)
{
	if(act != NULL)
	{
		app += act->name;
		app += ":\t";
		app += act->info;
		if(act->flags & ( ACT_CONTAINER | ACT_NOSUB | ACT_OPTARG ))
		{
			app += " [";
			if(act->flags & ACT_CONTAINER) app += " container";
			if(act->flags & ACT_NOSUB) app += " nosub";
			if(act->flags & ACT_OPTARG) app += " optarg";
			app += " ]";
		}
		if(act->argc)
		{
			app += "; parameters:";
			for(unsigned int i = 0; i < act->argc; ++i)
			{
				app += " ";
				app += act->arginfo[i];
				app += "(";
				switch(act->argt[i])
				{
					case ACT_INT:
						app += "integer";
					break;
					case ACT_FLOAT:
						app += "float";
					break;
					case ACT_STRING:
						app += "string";
					break;
					default:
					cerr << "[action] bad action argument type " << act->argt[i] << " most likely, this is a programming error that must be fixed" << endl;
			 }
			 app += ")";
			}
		}
	}
	else app += "unknown command";
}

bool action::parse(const string &command)
{
	clear(); //redundant?
	if(commando(command, fullname))
	{
		def = find_action_def(fullname, def_world);
		if(def != NULL)
		{
			/*
				definition found, try to parse command
				we have at def->argt def->argc integers stating the types
				so, split the command into the correct number of tokens and push them to the vectors
			*/
			
			vector<string> tokens;
						
			//optional arg means: either specify _all_ or _none_
			if(argo(command, tokens, def->argc) || def->flags & ACT_OPTARG)
			{
				if(!tokens.empty()) //if we indeed have some arguments, convert/store them
				{
					for(unsigned int i = 0; i < def->argc; ++i)
					{
						switch(def->argt[i])
						{
							case ACT_INT:
								ints.push_back(atol(tokens[i].c_str()));
							break;
							case ACT_FLOAT:
								floats.push_back(atof(tokens[i].c_str()));
							break;
							case ACT_STRING:
								strings.push_back(tokens[i]);
							break;
							default:
								cerr << "[action] bad action argument type " << def->argt[i] << " of argument " << i << " of command " << def->name << " - most likely, this is a programming error that must be fixed" << endl;
						}
					}
				}
				
				//now there may be 
				if(def->flags & ACT_CONTAINER)
				{
					if(!strings.empty()) //at least one string argument is required: the command for subaction
					{
						subaction = new action(def_world);
						subaction->comdat = comdat;
						subaction->swr = swr;
						if(subaction->parse(strings.back()))
						{
							if(! (subaction->def->flags & ACT_NOSUB) ) good = true;
							else strings.push_back("supposed subaction is not allowed to be subaction (not for script...)");
						}
						else strings.push_back("parsing of subaction failed");
						subaction->first = false;
						fullname += "/" + subaction->fullname;
						//even if we had errors, the subactions stay there (for diagnosis) until clear();
					}
					else strings.push_back("no command for subaction (argument string missing)");
				}
				else good = true;
			}
			else strings.push_back("unable to get (all) needed arguments");
			
			//parsed string stays here...
		}
		else strings.push_back("unknown command");
		
	}
	else strings.push_back("unable extract command name");

	if(good && def->flags & idmask)
	fullname += tstring().sprintf(" %i", ints.empty() ? -1 : ints[0]);

	return good;
}

/*
	reconstruct and print command string from action type and stored arguments
	
	this is kindof tricky with optional arguments...
	well, since I opted for all or none policy, this should apply
	
	note on subaction command string: when the last string of a container is not present, the subaction is called to append its command...
*/

bool action::print(string &command)
{
	char buffer[30];
	bool ret = true;
	if((def != NULL) && good)
	{
		command += def->name;
		//when this check doesn't guarantee that we won't try to access non-existent stuff, you messed too much with action internals!
		if
		(
			((ints.size()+strings.size()+floats.size()) >= def->argc)
			||
			((def->flags & ACT_CONTAINER) && (def->argc > 0) && (ints.size()+strings.size()+floats.size() == (def->argc - 1)))
		)
		{
			size_t icount = 0;
			size_t scount = 0;
			size_t fcount = 0;
			for(unsigned int i = 0; i < def->argc; ++i)
			{
				command += " ";
				switch(def->argt[i])
				{
					case ACT_INT:
						snprintf(buffer, 30, "%li", ints[icount++]);
						command += buffer;
					break;
					case ACT_FLOAT:
						snprintf(buffer, 30, "%0.3f", floats[fcount++]);
						command += buffer;
					break;
					case ACT_STRING:
						if((def->flags & ACT_CONTAINER) && (scount >= strings.size())) ret = subaction->print(command);
						else command += strings[scount];
						++scount;
					break;
					default:
					error("bad action argument type %i of argument %u of command %s - most likely, this is a programming error that must be fixed", def->argt[i], i, def->name);
				}
			}
		}
	}
	else ret=false;
	
	return ret;
}

void action::success(string &app)
{
	if(first) app += "[" + fullname + "] " + (good ? "success" : "error: ");
	if(!good)
	{
		if(!strings.empty()) app += strings.back();
		if(subaction != NULL){ app += " <- "; subaction->success(app); }
	}
}

void action::give_error(const string msg, string &app)
{
	app += "[" + fullname + "] " + "error: " + msg;
}

void action::msg(const string msg, string &app)
{
	app += "[" + fullname + "] " + msg;
}

void action::set_sub(action* newsub)
{
	subaction = newsub;
}

void action::begin(string &app)
{
	msg("+begin", app);
}

void action::end(string &app)
{
	msg("-end", app);
}

