#ifndef DMD_H_ACTION
#define DMD_H_ACTION

/*
	action: data structure for communicating orders and their execution
	
	part of DerMixD
	(c)2004-2008 Thomas Orgis, licensed under GPLv2
	
	actions do the parsing and de-parsing (creation of command string) on their own, provided they have a matching set of definitions
	actions transport arguments
	actions transport communication links
	actions can contain actions...
	actions are important
*/

#include "common.hxx"
#include "socket_writer_data.hxx"
#include "debug.hxx"

//constants: argument types and a limit
#define ACT_INT 1
#define ACT_STRING 2
#define ACT_FLOAT 3
#define ACT_MAX_ARG 5 //I thought I could use unspecified array in the struct... or pointer... seems not to be the case (?)

//flags for container, not allowed in container
#define ACT_CONTAINER 1
#define ACT_NOSUB 2
 //optional argument
#define ACT_OPTARG 4
//custom flags for external use
#define ACT_CUSTOM1 8
#define ACT_CUSTOM2 16
#define ACT_CUSTOM3 32
#define ACT_CUSTOM4 64
#define ACT_CUSTOM5 128
#define ACT_CUSTOM6 256
#define ACT_CUSTOM7 512
#define ACT_CUSTOM8 1024

/*
	there are two basic types of actions: single or container

	{ IN_EQ, "eq", "set all three eq factors", 4 , { "channel", "bass", "mid", "treble" }, { ACT_INT, ACT_FLOAT, ACT_FLOAT, ACT_FLOAT}, 0 }

	single action splits argument into argc pieces and puts them into strings, floats, ints, accordingly
	for container actions the last argument must be a string and this is not appended to strings but instead parsed again, creating the subaction.
	
	{ IN_SCRIPT, "script", "script an action triggered by an inchannel passing certain time/position, command is just any valid (and allowed) command with arguments", 3, { "channel", "time in seconds", "command" }, { ACT_INT, ACT_FLOAT, ACT_STRING }, ACT_CONTAINER|ACT_NOSUB }
	
	Hm, with this subaction pointer I guess I can endlessly contain actions in each other...
*/


//structure for definition of action syntax and storage
typedef struct action_def
{
	int id; // numerical ID
	const char* name; // "outload"
	const char* info;
	unsigned int argc; // 3
	const char* arginfo[ACT_MAX_ARG];
	int argt[ACT_MAX_ARG]; // {INT, STRING, STRING}
	int flags; // ACT_CONT, ACT_NOSUB 
} action_def;

//general tools for action definitions

//find definition of action in world with given command name
const action_def* find_action_def(string& name, const action_def* world);
//append list of all command names in this worls to app
void list_actions(const action_def* world, string& app);
//append action info and syntax help to app
void action_help(const action_def* act, string& app);
void action_allhelp(const action_def* world, vector<string*> &app);

//the action class itself: representation, storage, parsing

class action
{
	public:
	//unsigned char id; //type of action in some context
	const action_def* def; //the definition, including ID (subdef is filled for contained action)
	const action_def* def_world; //the world of definitions that is needed for parsing
	action* subaction; //the contained action
	bool first;
	bool good;
	int idmask;
	string fullname;

	//arguments for action
	vector<string> strings;
	vector<long>   ints; //enough? Better off_t in future with offsets (seeking) in mind?
	vector<double> floats;
	comm_data* comdat; //associated communication data (socketeer)
	socket_writer_data* swr; //associated socket_writer (output of resulting messages)

	//action(unsigned char i)
	action(const action_def* world)
	{
		construct("action");
		idmask = 0;
		subaction = NULL;
		comdat = NULL;
		swr = NULL;
		clear();
		def_world = world;
	};

	action(const action_def* world, const action_def* me)
	{
		construct("action");
		idmask = 0;
		subaction = NULL;
		comdat = NULL;
		swr = NULL;
		clear();
		def = me;
		def_world = world;
	};

	action()
	{
		construct("action");
		idmask = 0;
		subaction = NULL;
		comdat = NULL;
		swr = NULL;
		clear();
		def_world = NULL;
	};

	//just clear data
	void clear_data()
	{
		strings.clear();
		ints.clear();
		floats.clear();
	}

	//really clear
	void clear()
	{
		strings.clear();
		ints.clear();
		floats.clear();
		def = NULL;
		first = true;
		good = false;
		if(subaction != NULL)
		{
			delete subaction;
			subaction = NULL;
		}
		fullname.clear();
	};

	/*
		parse command and transform into the desired action,
		return true when definition found
	*/
	bool parse(const string &command);
	
	/*
		add command text that results in this action when parsed to given string
		idea is that this method should be useful when printing script actions...
	*/
	bool print(string &command);

	void success(string& app); //append error or success message(s) of myself and subactions to app
	void give_error(const string msg, string &app); //append given error message with [fullname] in front
	void msg(const string msg, string &app); //
	void begin(string &app);
	void end(string &app);
	
	~action()
	{
		destbegin("action");
		if(subaction != NULL) delete subaction;
		destend("action");
	}
	
	void set_sub(action* newsub); //just set subaction without further precautions... for now
};

#endif
