#ifndef DMD_H_AUDIO_BUFFER
#define DMD_H_AUDIO_BUFFER

/*
	audio_buffer: audio data storage
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include <math.h>
#include <limits.h>
#include "debug.hxx"

/*
	common struct for definition of audio formats	(sampling rate and number of channels)
	In theory one could support other channel counts than 2	but existing code relies on channels being 2 at some places!
*/
typedef struct audio_format
{
	unsigned long int rate;
	unsigned int channels;
} audio_format;


/*
	alias for base data type of audio samples
	This is flexible in principle but not in practice, esp. existing outputs all use short/16bit.
*/
typedef short audio_type;
// The absolute sample value is defined as the abs offset from the middle; which can be != 0 for unsigned samples.
#define AUDIO_MAX SHRT_MAX // 32767 or 32768 ... is this a philosophical decision?
#define audio_vol(sample)    ((double)(sample)/AUDIO_MAX)
#define audio_scale_vol(dsample)    ((dsample)/AUDIO_MAX)
#define audio_abs(sample)    ((audio_type)abs(sample))
#define audio_vol2dbfs(vol)  (log10(vol)*20)
#define audio_dbfs2vol(dbfs) pow(10, dbfs/20)

/*
	the audio buffer used throughout this project
	flexible in respect to different audio formats (including bits / audio_type)
	
	It provides mainly transparent allocation and reallocation of memory for the wanted number of samples
	together with administrative info as fill or virginity.
	
	...my first template class...
*/

//definition

template <class T, class PT, class RT>
class generic_audio_buffer
{
	public:
		RT channels;
		RT size;
		RT fill;
		PT ptr;
		generic_audio_buffer();
		generic_audio_buffer(unsigned int s);
		generic_audio_buffer(unsigned int c, unsigned int s);
		~generic_audio_buffer();
		unsigned int reallocate(unsigned int s);
		unsigned int reallocate(unsigned int c, unsigned int s);
		void zero();
		volatile bool virgin;
		size_t mysize(){ return sizeof(T)*channels*size; };
		size_t my_current_size(){ return sizeof(T)*channels*fill; };
		void printout();
		void print_text(ostream &wayout);
		void print_text_header(ostream &wayout);
		void swap(generic_audio_buffer<T,PT,RT> &willie);
};

//implementation

template <class T, class PT, class RT>
generic_audio_buffer<T,PT,RT>::generic_audio_buffer()
{
	construct("audio_buffer");
	ptr = NULL; size = 0; fill = 0; channels = 2;
	virgin = true;
}

template <class T, class PT, class RT>
generic_audio_buffer<T,PT,RT>::generic_audio_buffer(unsigned int s)
{
	construct("audio_buffer");
	ptr = NULL;  channels = 2; size = 0;
	reallocate(s);
}

template <class T, class PT, class RT>
generic_audio_buffer<T,PT,RT>::generic_audio_buffer(unsigned int c, unsigned int s)
{
	construct("audio_buffer");
	ptr = NULL;  channels = c; size = 0;
	reallocate(s);
}

template <class T, class PT, class RT>
generic_audio_buffer<T,PT,RT>::~generic_audio_buffer()
{
	destbegin("audio_buffer");
	if(ptr != NULL) delete[] ptr;
	destend("audio_buffer");
}

//change buffer size with memory reallocation
template <class T, class PT, class RT>
unsigned int generic_audio_buffer<T,PT,RT>::reallocate(unsigned int s)
{
	fill = 0;
	if(s != size)
	{
		if(ptr != NULL) delete[] ptr;
		if( (ptr = new T [s*channels]) != NULL ) size = s;
		else size = 0;
		virgin = false; //memory not initialized
	}
	return size;
}

//change buffer size and channels with memory reallocation
template <class T, class PT, class RT>
unsigned int generic_audio_buffer<T,PT,RT>::reallocate(unsigned int c, unsigned int s)
{
	fill = 0;
	if(s*c != size*channels) //if the needed memory size is different
	{
		if(ptr != NULL) delete[] ptr;
		if( (ptr = new T [s*c]) == NULL ) s = 0;
		virgin = false; //memory not initialized
	}
	channels = c;
	size = s;
	return size;
}

//sweep it clean
template <class T, class PT, class RT>
void generic_audio_buffer<T,PT,RT>::zero()
{
	//any benefit? seems not to be the case...
	//memset(const_cast<T*> (ptr), 0, mysize());
	for(size_t i = 0; i < size*channels; ++i) ptr[i] = 0;
	fill = 0;
	virgin = true;
}

//complete ascii printout
template <class T, class PT, class RT>
void generic_audio_buffer<T,PT,RT>::printout()
{
	print_text_header(cout);
	print_text(cout);
}

//print header for ascii text output
template <class T, class PT, class RT>
void generic_audio_buffer<T,PT,RT>::print_text_header(ostream &wayout)
{
	wayout << "#buffer printout" << endl << "#";
	for(RT ch = 0; ch < channels; ++ch)
	{
		if(ch) wayout << "\t";
		wayout << "channel" << ch;
	}
	wayout << endl;
}

//print buffer contents as ascii text
template <class T, class PT, class RT>
void generic_audio_buffer<T,PT,RT>::print_text(ostream &wayout)
{
	for(unsigned int i = 0; i < fill*channels; i+=channels)
	{
		for(RT ch = 0; ch < channels; ++ch)
		{
			if(ch) wayout << "\t";
			wayout << ptr[i+ch];
		}
		wayout << endl;
	}
}

//swapping of the memory of buffers with same properties
template <class T, class PT, class RT>
void generic_audio_buffer<T,PT,RT>::swap(generic_audio_buffer<T,PT,RT> &willie)
{
	PT tmp = ptr;
	ptr = willie.ptr;
	willie.ptr = tmp;
	
	RT tf = fill;
	fill = willie.fill;
	willie.fill = tf;
}

//instances

//with audio_type for normal read/write/store operation
typedef generic_audio_buffer<audio_type, volatile audio_type * volatile, volatile unsigned int> audio_buffer;
//more room with int for mixing (might add a float one for internal eq later...?)
typedef generic_audio_buffer<int, int*, unsigned int> mixer_buffer;
// another normal buffer but without thread safety
typedef generic_audio_buffer<audio_type, audio_type *, unsigned int> local_audio_buffer;

#endif
