#include "action.hxx"
#include "communicator_actions.hxx"

/*
	communicator_actions: the world of socketeer/mixer actions
	
	part of DerMixD
	(c)2006 Thomas Orgis, licensed under GPLv2
	
	This is the place where the main mixer actions - ro be received and prepared by socketeers - are defined.
	
	Have a look at the action_def struct in action.hxx to get an idea about how this works.
	Be aware of the ACT_MAX_ARG limit that currently says that an action definition can contain only 5 arguments!
	This limit can easily changed in action.hxx, though.
	
	Define an action as element of this list via
	
	{ <numerical id>, "command", "info text", <number of arguments>, { <list of argument names> }, { <list of numerical argument types> }, <flags> }
	
	The numerical id is used for easy identification of an action in both socketeers and mixer.
	Command and info text should be obvious...
	It may be beneficial to know the possible argument types: ACT_INT for integers, ACT_FLOAT for floats (doubles, actually), ACT_STRING for strings
	
	Also the last entry, flags, may need some explanation. As one may already have seen, you can put the OR of some flag constants in here to activate their flags.
	There are some flags for use inside the action's code:
	
	ACT_CONTAINER means that the action wraps over further action(s), the last argument must be a string that can be parsed into the subaction to be contained
	ACT_NOSUB prohibits placement of this action inside of a container
	ACT_OPTARG makes arguments optional; the action may be either parsed with setting all or none of the arguments.
	
	Then, there are some flags with meaning to socketeer/mixer:
	
	SOCK_LOCAL: handling of action inside socketeer without bothering the mixer
	SOCK_NOTIFY: triggered action reports back to socketeer, which has to notify the mixer afterwards
	SOCK_POST: this action needs postprocessing by socketeer before sending to mixer
	SOCK_IN: action on an inchannel (channel id as first integer argument required!)
	SOCK_OUT: action on an outchannel (channel id as first integer argument required!)
	SOCK_EXT: response to socket is not given by socketeer
	SOCK_ANS: a specialized answer is to be given (involding a return value...)
	
	Be sure not to omit the last entry with id 0! That one is the end marker... without it you'll get prompt segfaulting!
*/

const action_def in_start = { IN_START, "start", "start playback on inchannel", 1, { "channel" }, { ACT_INT }, SOCK_IN };
const action_def out_start = { OUT_START, "outstart", "start playback on outchannel", 1, { "channel" }, { ACT_INT }, SOCK_OUT };
const action_def in_script = { IN_SCRIPT, "script", "script an action triggered by an inchannel passing certain time/position, command is just any valid (and allowed) command with arguments", 3, { "channel", "time in seconds", "command" }, { ACT_INT, ACT_FLOAT, ACT_STRING }, ACT_CONTAINER|ACT_NOSUB|SOCK_IN|SOCK_POST };
const action_def in_nscript = { IN_SCRIPT, "nscript", "script an action triggered by an inchannel passing certain time/position, script will be executed n times (one for each triggering)", 4, { "channel", "n", "time in seconds",  "command" }, { ACT_INT, ACT_INT, ACT_FLOAT, ACT_STRING }, ACT_CONTAINER|ACT_NOSUB|SOCK_IN };

const action_def external_api[] =
{
	{ IN_VOLUME, "volume", "set inchannel volume factor", 2, { "channel", "volume" }, { ACT_INT, ACT_FLOAT }, SOCK_IN|SOCK_ANS },
	{ IN_BASS, "bass", "set inchannel bass eq factor", 2 , { "channel", "value" }, { ACT_INT, ACT_FLOAT}, SOCK_IN|SOCK_ANS },
	{ IN_MID, "mid", "set inchannel mid eq factor", 2 , { "channel", "value" }, { ACT_INT, ACT_FLOAT}, SOCK_IN|SOCK_ANS },
	{ IN_TREBLE, "treble", "set inchannel treble eq factor", 2 , { "channel", "value" }, { ACT_INT, ACT_FLOAT}, SOCK_IN|SOCK_ANS },
	{ IN_EQ, "eq", "set all three eq factors", 4 , { "channel", "bass", "mid", "treble" }, { ACT_INT, ACT_FLOAT, ACT_FLOAT, ACT_FLOAT}, SOCK_IN },
	{ IN_PAUSE, "pause", "pause inchannel", 1 , { "channel" }, { ACT_INT }, SOCK_IN },
	{ IN_SPEED, "speed", "set playback speed factor (1=normal)", 2 , { "channel", "value" }, { ACT_INT, ACT_FLOAT}, SOCK_IN|SOCK_ANS },
	{ IN_PITCH, "pitch", "increase/decrease playback speed factor", 2 , { "channel", "value" }, { ACT_INT, ACT_FLOAT}, SOCK_IN|SOCK_ANS },
	{ IN_STOP, "stop", "stop playback on inchannel", 1, { "channel" }, { ACT_INT }, SOCK_IN },
	in_start,
	{ IN_SEEK, "seek", "absolute seek on inchannel", 2 , { "channel", "position in seconds" }, { ACT_INT, ACT_FLOAT }, SOCK_NOTIFY|SOCK_IN|SOCK_ANS },
	{ IN_RSEEK, "rseek", "relative (from current position) seek on inchannel", 2 , { "channel", "offset in seconds" }, { ACT_INT, ACT_FLOAT }, SOCK_NOTIFY|SOCK_IN|SOCK_ANS },
	{ IN_BIND, "bind", "bind input channel to output channel", 2 , { "inchannel", "outchannel" }, { ACT_INT, ACT_INT }, SOCK_IN },
	{ IN_UNBIND, "unbind", "release bond between input channel and output channel", 2 , { "inchannel", "outchannel" }, { ACT_INT, ACT_INT }, SOCK_IN },
	{ OUT_PAUSE, "outpause", "pause outchannel", 1 , { "channel" }, { ACT_INT }, SOCK_OUT },
	{ OUT_STOP, "outstop", "stop playback on outchannel", 1, { "channel" }, { ACT_INT }, SOCK_OUT },
	out_start,
	in_script,
	in_nscript,
	{ IN_SHOWSCRIPT, "showscript", "show the script commands for programmed actions on an inchannel", 1, { "the channel" }, { ACT_INT }, SOCK_EXT|ACT_NOSUB|SOCK_IN },
	{ IN_CLEARSCRIPT, "delscript", "remove scripting actions from inchannel", 1, { "channel" }, { ACT_INT }, SOCK_IN },
	{ IN_FOLLOW, "follow", "let one inchannel follow another (start folower gaplessy after leader stopped)", 2 , { "leader", "follower" }, { ACT_INT, ACT_INT }, SOCK_IN },
	{ IN_NOFOLLOW, "nofollow", "release the followership on a leader", 1 , { "leader" }, { ACT_INT }, SOCK_IN },
	{ GETSTAT, "getstat", "give status info in mixplayd-like format (please do not use... may change/vanish completely)", 0, { }, { }, ACT_NOSUB|SOCK_EXT },
	{ GETSTAT, "g", "short for getstat", 0, { }, { }, ACT_NOSUB|SOCK_EXT },
	{ FULLSTAT, "fullstat", "give full status info on all in- and outchannels (several lines)", 0, { }, { }, ACT_NOSUB|SOCK_EXT },
	{ FULLSTAT, "f", "short for fullstat", 0, { }, { }, ACT_NOSUB|SOCK_EXT },
	{ SAY, "say", "just say something (I put [saying] in front); useful when timed somehow", 1, { "the message" }, { ACT_STRING }, 0 },
	{ IN_WATCH, "watch", "watch inchannel (get messages on events and ongoing playback)", 1, { "channel" }, {ACT_INT}, SOCK_IN }, //better nosub?
	{ IN_UNWATCH, "unwatch", "stop watching inchannel (see watch)", 1, { "channel" }, { ACT_INT }, SOCK_IN }, //nosub?
	//this is to be translated into IN_LOAD+IN_START by socketeer!
	{ IN_LOAD, "play", "load track and start playback on inchannel", 2, { "channel", "track" }, { ACT_INT, ACT_STRING }, ACT_NOSUB|SOCK_POST|SOCK_NOTIFY|SOCK_IN|SOCK_PATH },
	{ IN_LOAD, "load", "load track on inchannel (be prepared for starting it)", 2, { "channel", "track" }, { ACT_INT, ACT_STRING }, ACT_NOSUB|SOCK_NOTIFY|SOCK_IN|SOCK_PATH },
	{ IN_INLOAD, "inplay", "load track and start playback on inchannel", 3, { "channel", "driver", "track" }, { ACT_INT, ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_POST|SOCK_NOTIFY|SOCK_IN|SOCK_PATH },
	{ IN_INLOAD, "inload", "load track on inchannel (be prepared for starting it)", 3, { "channel", "driver", "track" }, { ACT_INT, ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_NOTIFY|SOCK_IN|SOCK_PATH },
	{ IN_SCAN, "scan", "scan input properties", 2, { "channel", "list of properties" }, { ACT_INT, ACT_STRING }, ACT_NOSUB|SOCK_NOTIFY|SOCK_IN },
	{ OUT_LOAD, "outload", "load resource on outchannel with specified driver/device (prepared for activity)", 3, { "channel", "driver", "resource" }, { ACT_INT, ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_NOTIFY|SOCK_OUT|SOCK_PATH },
	//this is to be translated into OUT_LOAD+OUT_START by socketeer!
	{ OUT_LOAD, "outplay", "load resource on outchannel and start playback", 3, { "channel", "driver", "resource" }, { ACT_INT, ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_POST|SOCK_NOTIFY|SOCK_OUT|SOCK_PATH },
	{ IN_LENGTH, "length", "determine exact length of track (stops channel, decodes track till end, seeks back to where it left off)", 1, { "channel" }, { ACT_INT }, SOCK_NOTIFY|SOCK_IN|SOCK_ANS },
	{ IN_PREREAD, "preread", "read through a file once (to have it cached by file system/kernel)", 1, { "file" }, { ACT_STRING }, SOCK_LOCAL|SOCK_PATH|ACT_NOSUB },
	//hm, optional argument... socketeer can handle with two action parse calls
	{ ADD_IN, "addin", "add input channel (with optional nick name)", 1, { "nick name" }, { ACT_STRING }, ACT_NOSUB|ACT_OPTARG|SOCK_POST },
	{ REM_IN, "remin", "remove input channel", 1, { "id" }, { ACT_INT }, ACT_NOSUB },
	{ ADD_OUT, "addout", "add output channel (with optional nick name)", 1, { "nick name" }, { ACT_STRING }, ACT_NOSUB|ACT_OPTARG|SOCK_POST },
	{ REM_OUT, "remout", "remove output channel", 1, { "id" }, { ACT_INT }, ACT_NOSUB },
	{ ID, "id", "print my full identification", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL },
	{ CLOSE, "close", "close current connection", 0, { }, { }, SOCK_POST },
	{ FADEOUT, "fadeout", "dummy to remind you to do custom fading via script actions", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL|ACT_OPTARG },
	{ IN_VOLUME, "vol", "set inchannel volume in percent", 2, { "channel", "volume" }, { ACT_INT, ACT_FLOAT}, SOCK_POST|SOCK_IN|SOCK_ANS }, //make it percent!
	{ IN_BUFFER, "buffer", "load audio into mixer buffer (this includes zeroscan)", 1, { "channel" }, { ACT_INT }, SOCK_IN },
	{ SLEEP, "sleep", "let me sleep until any client wants something", 0, { }, { }, 0 },
	{ QUIT, "shutdown", "lay down and die gracefully", 0, { }, { }, SOCK_POST },
	{ HELP, "help", "give some info / usage pattern on a command", 1, { "the command" }, { ACT_STRING }, ACT_OPTARG|SOCK_LOCAL },
	{ SHOWAPI, "showapi", "show the whole API (list help for all commands)", 0, { }, { }, SOCK_LOCAL|ACT_NOSUB },
	{ THREADSTAT, "threadstat", "list spawned threads", 0, { }, { }, ACT_NOSUB },
	{ LS, "ls", "list file/directory", 1, { "dir/file" }, { ACT_STRING }, ACT_NOSUB|SOCK_LOCAL|ACT_OPTARG },
	{ CD, "cd", "change directory (for client actions)", 1, { "dir" }, { ACT_STRING }, ACT_NOSUB|SOCK_LOCAL },
	{ PWD, "pwd", "print current working directory", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL },
	{ SPY, "spy", "spy on client communication", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL },
	{ UNSPY, "unspy", "stopy spying on client communication", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL },
	{ SHOWID, "showid", "toggle showing of channel id in responses to channel commands", 1, { "1(on) or 0(off)" }, { ACT_INT }, ACT_NOSUB|SOCK_LOCAL },
	{ PEER, "peer", "send a message to another peer (client).", 3, { "your peer name", "peer's peer name", "message" }, { ACT_STRING, ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_LOCAL }, 
	{ ADDPEER, "addpeer", "add a peer entry, registering for messages", 2, { "peer name", "description" }, { ACT_STRING, ACT_STRING }, ACT_NOSUB|SOCK_LOCAL },
	{ REMPEER, "rempeer", "remove a peer entry", 1, { "peer name" }, { ACT_STRING }, ACT_NOSUB|SOCK_LOCAL },
	{ SHOWPEERS, "showpeers", "show a list of peers with optional description", 0, { }, { }, ACT_NOSUB|SOCK_LOCAL },
	{ 0, "", "", 0, { }, { }, 0 }
};

const action_def in_notify = { NOTIFY, "<internal>", "notification", 0, { }, { }, ACT_CONTAINER|SOCK_IN };
const action_def out_notify = { NOTIFY, "<internal>", "notification", 0, { }, { }, ACT_CONTAINER|SOCK_OUT };
const action_def notify = { NOTIFY, "<internal>", "notification", 0, { }, { }, ACT_CONTAINER };
