/*
	mixer_data: data types and defines for mixer 
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
*/
#include "common.hxx"

#include "netcomm.hxx"

#include "audio_buffer.hxx"
#include "audio_functions.hxx"

#include "input.hxx"

#include "output.hxx"

#include "socket_writer_data.hxx"
#include "action.hxx"
#include "mixer_data.hxx"
#include "mixer_tools.hxx"
#include "communicator_actions.hxx"

//#define DEBUG_SCRIPT
//#define DEBUG_MIXER2
//#define DEBUG_RESET

/* load errors:
	0: good
	1: file error
	2: type unknown
	3: device failure
*/

//an input channel

//get all actions falling in upcoming chunk, sorted ascending in time 
bool channel::getactions(unsigned int chunksize, vector<action*> &nowscript)
{
	double now,then;
	now_and_then(chunksize,now,then);
	//only search when the window is open
	if
	(
		(!script.empty())
		&& (script.front().time < then)
		&& (script.back().time >= now)
	)
	{
		xdebug("[mixer ch %s] really looking for action", name.c_str());
		for(vector<timeaction>::iterator t = script.begin(); t != script.end(); ++t)
		{
			if((*t).time >= now)
			{
				if((*t).time < then)
				{
					if((*t).count > 0 && --(*t).count == 0)
					{
						nowscript.push_back((*t).act);
						script.erase(t);
						--t; //after that ++t should point to the next item... hopefully
					}
					else //make a copy so that action can be executed again
					{
						action* copyact = new action;
						*copyact = *(*t).act;
						nowscript.push_back(copyact);
					}
				}
				else break; //since script is sorted ascending, there should be nothing more to be found since time >= then
			}
		}
	}
	return (!script.empty());
}

bool channel::getpostactions(vector<action*> &nowscript)
{
	xdebug("[mixer ch %s] really looking for postaction", name.c_str());
	for(vector<timeaction>::iterator t = postscript.begin(); t != postscript.end(); ++t)
	{
		if((*t).count > 0 && --(*t).count == 0)
		{
			nowscript.push_back((*t).act);
			postscript.erase(t);
			--t; //after that ++t should point to the next item... hopefully
		}
		else //make a copy so that action can be executed again
		{
			action* copyact = new action;
			*copyact = *(*t).act;
			nowscript.push_back(copyact);
		}
	}
	return (!postscript.empty());
}

//add action to internal script in an ascending sorted way
vector<timeaction>::size_type channel::addaction(const double time, const long count,  action* act)
{
	timeaction ta;
	ta.time = time;
	ta.count = count;
	ta.act = act;
	if(ta.count == 0) ta.count = 1;
	#ifdef DEBUG_SCRIPT
	cerr << "[channel " << name << "] adding action of type " << (int) act->id << " timed at " << time << "s" << endl;
	#endif
	//think about reverse iterator!
	vector<timeaction>::iterator t;
	for(t = script.begin(); t != script.end(); ++t)
	{
		if( (*t).time > time ) break;
	}
	script.insert(t, ta);
	#ifdef DEBUG_SCRIPT
	fprintf(stderr, "[inchannel] added action pointer %p\n", (void*) script.back().act);
	#endif
	return script.size();
}

vector<timeaction>::size_type channel::addpostaction(const double time, const long count,  action* act)
{
	timeaction ta;
	ta.time = time;
	ta.count = count;
	ta.act = act;
	if(ta.count == 0) ta.count = 1;
	#ifdef DEBUG_SCRIPT
	cerr << "[channel " << name << "] adding postaction of type " << (int) act->id << " timed at (virtual) " << time << "s" << endl;
	#endif
	//think about reverse iterator!
	vector<timeaction>::iterator t;
	for(t = postscript.begin(); t != postscript.end(); ++t)
	{
		// smaller times (larger absolute value) are closer to te end
		// time -2s is scheduled before -1s
		if( (*t).time > time ) break;
	}
	postscript.insert(t, ta);
	#ifdef DEBUG_SCRIPT
	fprintf(stderr, "[inchannel] added action pointer %p\n", (void*) postscript.back().act);
	#endif
	return postscript.size();
}



void channel::delscript()
{
	for(vector<timeaction>::iterator t = script.begin(); t != script.end(); ++t)
	{
		delete (*t).act;
	}
	script.clear();
	for(vector<timeaction>::iterator t = postscript.begin(); t != postscript.end(); ++t)
	{
		delete (*t).act;
	}
	postscript.clear();
}

/*
	The socket of dead_writer is being closed:
	Set the writer in all script actions that may want to produce output to NULL so that there is no attempt to use the long gone dead_writer in the future.
*/
void channel::script_nirvana(socket_writer_data* dead_writer)
{
	for(vector<timeaction>::iterator t = script.begin(); t != script.end(); ++t)
	if((*t).act->swr == dead_writer) (*t).act->swr = NULL;

	for(vector<timeaction>::iterator t = postscript.begin(); t != postscript.end(); ++t)
	if((*t).act->swr == dead_writer) (*t).act->swr = NULL;
}

//reconstruct script commands
void inchannel::get_script_commands(vector<string*> &list)
{
	action scriptac(external_api);
	scriptac.good = true;
	scriptac.floats.push_back(0);
	scriptac.ints.push_back(data.id);
	scriptac.ints.push_back(1);
	for(vector<timeaction>::iterator t = script.begin(); t != script.end(); ++t)
	{
		string* bully = new string;
		scriptac.subaction = (*t).act;
		scriptac.floats[0] = (*t).time;
		scriptac.ints[1] = (*t).count;
		scriptac.def = (*t).count > 1 ? &in_nscript : &in_script;
		scriptac.print(*bully);
		list.push_back(bully);
	}
	for(vector<timeaction>::iterator t = postscript.begin(); t != postscript.end(); ++t)
	{
		string* bully = new string;
		scriptac.subaction = (*t).act;
		scriptac.floats[0] = (*t).time;
		scriptac.ints[1] = (*t).count;
		scriptac.def = (*t).count > 1 || (*t).count < 0 ? &in_nscript : &in_script;
		scriptac.print(*bully);
		list.push_back(bully);
	}
	scriptac.subaction = NULL; //don't delete action on destruct!
}


inchannel::inchannel(int mid, const string& namen, sem_t* inns)
{
	construct("inchannel");
	working = false;
	offline = false;
	data.insem = inns;
	data.id = mid;
	follower = NULL;
	leader = NULL;
	device = NULL;
	name = namen;
	startval[0] = 0;
	startval[1] = 0;
	buffering = false;
	hadseconds = 0;
	hadstatus = NOTHING;
	speed = 1; volume = 1; status = IDLE; 
	data.format.rate = audio_rate;
	data.format.channels = audio_channels;
	data.at_end = true;
	data.bass = 1;
	data.mid = 1;
	data.treble = 1;
	resize_buffer(buffer_samples);
	#ifdef DEBUG_MIXER2
	cout << "new inchannel with id " << data.id << endl;
	#endif
}

void inchannel::nodev()
{
	if(device != NULL) { delete device; device = NULL; }
	sub_thread_container = NULL;
	data.reset();
	status = NODEV;
}


unsigned char inchannel::load(string &resource, comm_data *comdat, const string type)
{
	unsigned char errcode = 0;
	data.reset();
	hadseconds = 0;
	hadstatus = NOTHING;
	offline = false;
	if(type == "nodev")
	{
		offline = true;
		debug("[mixer ch %i] just clearing device on request", data.id);
		nodev();
		status = IDLE;
		if(comdat != NULL) sem_post(comdat->waitress);
		return 0;
	}

	audiodata* ad = get_audiodata(resource, type);
	debug("[mixer ch %i] get_audiodata done", data.id);
	if(ad->exists)
	{
		debug("[mixer ch %i] got some audiodata for %s", data.id, resource.c_str());
		if(ad->type != "")
		{
			debug("[mixer ch %i] it is of known type %s", data.id, ad->type.c_str());
			//can I reuse a device?
			if( (device != NULL) && (strcasecmp(device->type->c_str(),ad->type.c_str()) != 0) )
			{
				debug("[mixer ch %i] deleting input device (address: %p) because its type (%s) does not match the new titles type (%s)", data.id, (void*) device, device->type->c_str(), ad->type.c_str());
				nodev();
			}
			//no, I'll have to create one
			if(device == NULL)
			{
				debug("[mixer ch %i] going to create a new input device", data.id);
				device = new_input_device(ad->type, data);
				sub_thread_container = device;
				debug("[mixer ch %i] new indevice created on channel", data.id);

				status = IDLE;
			}
			else debug("[mixer ch %i] reusing input device (address: %p)", data.id, (void*) device);

			if( (device == NULL) || (!device->online()) ) 
			{
				error("[mixer ch %i] This device is not online, removing it -- this load failed.", data.id);
				comdat->errtext += (device == NULL ? "no device for media type " + ad->type : (device->problem == "" ? "device failed to init" : device->problem)) + "; ";
				nodev();
				errcode = 3;
			}
			else
			{
				debug("[mixer ch %i] online! setting data", data.id);
				//now we can load... these settings may be overridden by load!
				data.specials = ad->specials;
				data.format.rate = ad->format.rate;
				data.format.channels = ad->format.channels;
				data.length = ad->length;
				debug("[mixer ch %i] loading file %s", data.id, resource.c_str());
				device->load(comdat, resource);
				status = LOADING;
				debug("[mixer ch %i] load issued...", data.id);
			}
		}
		else{ errcode = 2; comdat->errtext += "unknown media type; "; }
	}
	else{ errcode = 1; comdat->errtext += "unable to determine basic media information: " + ad->problem; }

	delete ad;

	debug("[mixer ch %i] load method finished", data.id);
	return errcode;
}

bool inchannel::device_sane()
{
	if(device != NULL)
	{
		if(device->online()) return true;
		else { nodev(); return false; }
	}
	else return false;
}

bool inchannel::resize_buffer(unsigned int s)
{
	if(buffer.reallocate(data.format.channels,s) == 0){ status = DEAD;	return false; }
	else return true;
}

//prepare for re-read of invalidated buffer
void inchannel::reset_buffer()
{
	#ifdef DEBUG_RESET
	cout << "[reset_buffer " << data.id << "] called" << endl;
	#endif
	if(device != NULL)
	{
		#ifdef DEBUG_RESET
		cout << "[reset_buffer " << data.id << "] rewind from " << data.position << " to " << ( data.position < buffer.fill ) ? 0 : data.position - buffer.fill) << endl;
		#endif
		if(data.position < buffer.fill) device->sampleseek(NULL, 0);
		else device->sampleseek(NULL, data.position - buffer.fill);
	}
	buffer.fill = 0;
	if(leader != NULL) buffering = true;
}

inchannel::~inchannel()
{
	destbegin("inchannel");
	if( device != NULL ) delete device;
	delscript();
	destend("inchannel");
}

void inchannel::now_and_then(unsigned int chunksize, double &now, double &then)
{
	//beware of buffering! (did I get it here?)
	now = seconds(&data.format, data.position-buffer.fill);
	then = seconds(&data.format, (unsigned long) (data.position+speed*chunksize-buffer.fill));
}

//an outchannel

outchannel::outchannel(int mid, const string& namen)
{
	construct("outchannel");
	data.id = mid;
	name = namen;
	device = NULL;
	status = NODEV; //for now...
	working = false;
	offline = false;
	data.format.rate = audio_rate; //correct place??
	data.format.channels = audio_channels;
	#ifdef DEBUG_MIXER2
	cout << "new outchannel with id " << data.id << endl;
	#endif
}

bool outchannel::device_sane()
{
	if(device != NULL)
	{
		if(device->online()) return true;
		else { nodev(); return false; }
	}
	else return false;
}

unsigned char outchannel::load(const string type, const string resource, comm_data *comdat)
{
	unsigned char errcode = 0;
	data.reset();

	offline = false;
	if(type == "nodev")
	{
		offline = true;
		debug("[mixer ch %i] just clearing device on request", data.id);
		nodev();
		status = IDLE;
		if(comdat != NULL) sem_post(comdat->waitress);
		return 0;
	}
	//can I reuse a device?
	if( (device != NULL) && (strcasecmp(device->type().c_str(),type.c_str()) != 0) ) 
	{
		#ifdef DEBUG_MIXER2
		cout << "[mixer] deleting output device (address: " << (void*) device << ") on channel " << data.id << " because its type (" << device->type() << ") does not match the wanted type (" << type << ")" << endl;
		#endif
		nodev();
	}
	//no, I'll have to create one
	if(device == NULL)
	{
		#ifdef DEBUG_MIXER2
		cout << "[mixer] going to create a new output device on ch" << data.id << endl;
		#endif
		device = new_output_device(type, data);
		sub_thread_container = device;
		status = IDLE;

		#ifdef DEBUG_MIXER2
		cout << "[mixer] new outdevice created on channel" << data.id << endl;
		#endif
	}
	#ifdef DEBUG_MIXER2
	else cout << "[mixer] reusing output device (address: " << (void*) device << ") on channel " << data.id << endl;
	#endif
	if( (device == NULL) || (!device->online()) ) 
	{
		#ifdef DEBUG_MIXER2
		cout << "[mixer] this device is not online, removing it; this load failed\n";			
		#endif
		comdat->errtext += (device == NULL ? "no device for type " + type : (device->problem == "" ? "device failed to init" : device->problem)) + "; ";
		nodev();
		errcode = 1;
	}
	else
	{
		//some intelligent parsing of the resource concerning special syntax in the output device?
		#ifdef DEBUG_MIXER2
		cout << "[mixer] online! setting data\n";			
		#endif
		//now we can load...
		#ifdef DEBUG_MIXER2
		cout << "[mixer] loading resource " << resource << " on outchannel " << data.id << endl;
		#endif
		device->load(comdat, resource);
		status = LOADING;
		#ifdef DEBUG_MIXER2
		cout << "[mixer] load issued... errcode (still?) = " << (int) comdat->errcode << endl;
		#endif
	}
	return errcode;
}

void outchannel::nodev()
{
	if(device != NULL) { delete device; device = NULL; }
	sub_thread_container = NULL;
	data.reset();
	status = NODEV;
}

outchannel::~outchannel()
{
	destbegin("outchannel");
	if( device != NULL ) delete device;
	#ifdef DEBUG_MEM
	cout << "destruct outchannel: device gone" << endl;
	#endif
	delscript();
	destend("outchannel");
}


//script helper
void outchannel::now_and_then(unsigned int chunksize, double &now, double &then)
{
	now = seconds(&data.format, data.position);
	then = seconds(&data.format, data.position+chunksize);
}
