/*
	output_base: partial implementation of the virtual output_device class
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
	
	common stuff including mixer interaction and the multithreaded double-buffering framework
	
	This is a _very_ important brick in the wall of sound!
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "thread_container.hxx"

void output_device::prebuild()
{
	good_writer_error = 0;
	on_line = false;
	active = false;
	playing = false;
	have_sems = false;
	writer = NULL;
}

void output_device::build(output_data* odi, unsigned int bufsize, void* (*thewriter)(void*))
{
	#ifdef DEBUG_OUTPUT
	cout << "[output_device] create an output" << endl;
	#endif
	//initial setup for instance
	prebuild();
	data = odi;
	wdat.id = data->id;
	writer = thewriter;
	data->format.rate = audio_rate;
	data->format.channels = audio_channels;
	
	if(special_init())
	{
		if(resize_buffer(bufsize))
		{
			#ifdef DEBUG_OUTPUT
			cerr << "[output_device " << data->id << "] initializing semaphores " << (void*) &wdat.insem << " and " << (void*) &wdat.outsem << endl;
			#endif
			if(writer != NULL) //only create thread when we have a thread worker
			{
				if((sem_init(&wdat.insem, 0 , 0) != 0) || (sem_init(&wdat.outsem, 0 , 0) != 0)) 
				{
					problem = "Cannot initialize semaphore!";
				}
				else
				{
					have_sems = true;
					#ifdef DEBUG_OUTPUT
					int insemval, outsemval;
					sem_getvalue(&wdat.insem,&insemval);
					sem_getvalue(&wdat.outsem,&outsemval);
					cerr << "[output_device " << data->id << "] initialized semaphores " << (void*) &wdat.insem << "(value=" << insemval << ") and " << (void*) &wdat.outsem << "(value=" << outsemval << ")" << endl;
					#endif
					if(pthread_create(&wthread, NULL, writer, (void*) &wdat) == 0)
					{
						add_thread(wthread, "output writer", &wdat);
						on_line = true;
					}
				}
			}
			else on_line = true;
		}
	}
	#ifdef DEBUG_OUTPUT
	cerr << "[output_device " << data->id << "] build finished... on_line=" << on_line << endl;
	#endif
}

void output_device::destroy()
{
	#ifdef DEBUG_OUTPUT
	cout << "[output_device " << data->id << "] destruct" << endl;
	#endif
	
	deactivate(); //stop writer, halt device...
	
	if(writer != NULL)
	{
		//cancelling the thread before destructing any special stuff, although the thread should be sleeping anyway
		//on_line now has a concrete meaning: there is a thread; if not, it would be fatal (segfault) to try to cancel it with NPTL
		//there is memory lost on pthread_create... maybe the join is necessary here?
		//that doesn't change a thing ... can I really believe valgrind (or the old linuxthreads) here?
		//check that again for realness
		if(on_line)
		{
	  	if(sem_cancel(wthread, &wdat.insem) != 0) cerr << "[output_device " << data->id << "] error cancelling writer thread / thread not there" << endl;
			else pthread_join(wthread, NULL);
		}
		#ifdef DEBUG_OUTPUT
		else cout << "[output_device " << data->id << "] writer thread finished (or wasn't there anyway)" << endl;
		#endif
		on_line = false;
		special_exit(); //destruct any special stuff (device handles, config structures,...)
		if(have_sems)
		{
			#ifdef DEBUG_OUTPUT
			cout << "[output_device " << data->id << "] destroying semaphores" <<  endl;
			#endif
			sem_destroy(&wdat.insem);
			#ifdef DEBUG_OUTPUT
			cout << "[output_device " << data->id << "] insem destroyed" <<  endl;
			#endif
			sem_destroy(&wdat.outsem);
			#ifdef DEBUG_OUTPUT
			cout << "[output_device " << data->id << "] outsem destroyed" <<  endl;
			#endif
		}
	}
}

bool output_device::activate()
{
	if(!active) active = special_activate();
	return active;
}

void output_device::deactivate()
{
	#ifdef DEBUG_OUTPUT
	cout << "[output_device] deactivate()\n"; 	
	#endif
	if(active)
	{
		endgame();
		special_deactivate();
		active = false;
	}
}

void output_device::special_write()
{
	//just placeholder... really needed?
}

//this should return an error to the caller...
void output_device::play()
{
	#ifdef DEBUG_OUTPUT2
	cout << "[output_device] play()" << endl; 
	#endif
	
	if(active && buffer.fill != 0)
	{
		endgame(); //wait for last chunk to finish so that we can use it now
		//switch buffer, signal thread
		//_the_ buffer has been written to... need to switch ptr with the writer's buffer
		buffer.swap(wdat.buffer);
		#ifdef DEBUG_OUTPUT
		cout << "[output_device] play: posting" << endl; 
		#endif
		if(writer != NULL) sem_post(&wdat.insem); //wake the writer
		else special_write();
		//sem_wait(&wdat.outsem); //additional wait for debugging
		buffer.zero(); //be ready for new data
		playing = true;
	}
	#ifdef DEBUG_OUTPUT2
	cout << "[output_device] play() done" << endl; 
	#endif
}

void output_device::load(comm_data* cd, const string& filename)
{
	deactivate();

	debug("[ch%i] load: %s", data->id, filename.c_str());
	data->filename = filename;
	data->position = 0;

	// do _not_ activate it here, that's the job of activate!
	buffer.zero();
	cd->errcode = 0;

	sem_post(cd->waitress);
}

void output_device::pause()
{
	if(active)
	{
		endgame(); //really???
		special_pause();
	}
}

bool output_device::online()
{
	return on_line;
}

bool output_device::is_active()
{
	return active;
}

bool output_device::is_playing()
{
	return playing;
}

bool output_device::resize_buffer(unsigned int s)
{
	bool good = true;
	endgame(); //don't touch buffers while playing...
	bool was_active = active;
	deactivate(); //the buffer may affect the downlevel stuff
	
	if((buffer.reallocate(data->format.channels,s) == 0) || (wdat.buffer.reallocate(data->format.channels,s) == 0))
	{
		good = false;
	}
	
	#ifdef DEBUG_OUTPUT
	cout << "[output_device] resize_buffer: buffers now at " << (void*) buffer.ptr << " and " << (void*) wdat.buffer.ptr << " good: " << good << endl; 
	#endif
	
	buffer.zero();
	wdat.buffer.zero();
	
	//re-activate downlevel device if possible and wanted
	if(good && was_active){ activate(); good = active; }
	
	return good;
}

void output_device::endgame()
{
	if(playing)
	{
		if(writer != NULL)
		{
			#ifdef DEBUG_OUTPUT
			int semval;
			sem_getvalue(&wdat.outsem, &semval);
			cout << "[output_device " << data->id << "] waiting for finish of writer at " << &wdat.outsem << "(value=" << semval << ")"<< endl; 
			#endif
			//while(sem_trywait(&wdat.outsem) == EAGAIN){ sleep(1); cout << "[output_device " << data->id << "] and waiting..." << endl; }
			safe_sem_wait(&wdat.outsem);
		}
		data->position += wdat.buffer.fill;
		wdat.buffer.fill = 0; //now emtpy, being filled by mixer after swap
		playing = false;
		if(wdat.error != good_writer_error) //this has to be communicated upwards!
		{
			deactivate();
		}
	}
}
